'''pdfcombine
    Combine several PDFs to a single PDF.

    By default a PostScript script is used to set the meta-data of the output PDF-file.
    In particular, the output PDF gets a table of contents with bookmarks to the first page
    of each input 'document' and the input filename at title.
    To customise these titles and add meta-data use a YAML input file and/or
    customise the default PostScript script.

    Manipulate PostScript script:

    *   Customise:

            title
                Set title of the output PDF.

            author
                Set the author of the output PDF.

            no-bookmarks
                Switch-off bookmarks added for each 'document'.

            add-ps
                Add lines of PostScript code to the automatically generated script.

    *   Manually specify:

            ps
                Set PostScript script (overwrites automatically generated script).

    *   Suppress:

            no-ps
                Switch-off the use of a PostScript script.

    To include custom bookmarks a YAML input file can be used, e.g.::

        files:
            - file: 1.pdf
              title: First file
            - file: 2.pdf
              title: Second file

        openleft: True
        title: Binder
        author: Tom de Geus
        output: binder.pdf
        ...

    As observed the `files` field contains all input files (in the correct order) and the
    bookmark titles. In addition, any of the command-line options (long name without `--`)
    can be included. Note that specifying them will overwrite the corresponding
    command-line option. To use with automatic bookmarks (i.e. filenames),
    the above input file can be shortened to::

        files:
            - 1.pdf
            - 2.pdf

        openleft: True
        title: Binder
        author: Tom de Geus
        output: binder.pdf

Usage:
    pdfcombine [options] <files>...

Options:
    -y, --yaml              Read input files (and settings) from a YAML-file.
        --openleft          Enforce that each 'chapter' starts on an even page.
        --openright         Enforce that each 'chapter' starts on an odd page.
        --title=<arg>       Set the title of the output PDF.
        --author=<arg>      Set the author of the output PDF.
        --no-bookmarks      Do include bookmarks to the first page of each document.
        --add-ps=<arg>      Add commands to the generated PostScript script.
        --ps=<arg>          Overwrite the automatically generated PostScript script.
        --no-ps             Do not run any PostScript script (to edit meta-data).
    -o, --output=<arg>      Name of the output file. [default: binder.pdf]
    -f, --force             Force overwrite of existing output-file.
    -s, --silent            Do not print any progress.
        --verbose           Verbose all commands.
    -h, --help              Show help.
        --version           Show version.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/pdfcombine
'''

import docopt
import click
import os
import sys

from .. import __version__
from .. import combine


def error(text):
    r'''
Command-line error: show message and quit with exit code "1"
    '''

    print(text)
    sys.exit(1)


def read_yaml(files):
    r'''
Read YAML file.
    '''

    import yaml

    if len(files) != 1:
        error('All input PDFs need to be specified in the YAML input')

    filename = files[0]

    if not os.path.isfile(filename):
        error('"{0:s} does not exist'.format(filename))

    return yaml.load(open(filename, 'r').read(), Loader=yaml.FullLoader)


def main():
    r'''
Main program.
    '''

    args = docopt.docopt(__doc__, version=__version__)

    files = args['<files>']
    output = args['--output']
    openleft = args['--openleft']
    openright = args['--openright']
    ps = args['--ps']
    add_ps = args['--add-ps']
    meta = not args['--no-ps']
    bookmarks = not args['--no-bookmarks']
    title = args['--title']
    author = args['--author']
    verbose = args['--verbose']
    silent = args['--silent']

    if args['--yaml']:

        info = read_yaml(files)

        if 'files' in info:
            if len(info['files']) > 0:
                if type(info['files'][0]) == dict:
                    files = [i['file'] for i in info['files']]
                    bookmarks = [i['title'] for i in info['files']]
                else:
                    files = [i for i in info['files']]

        if 'output' in info:
            output = info['output']

        if 'openleft' in info:
            openleft = info['openleft']

        if 'openright' in info:
            openright = info['openright']

        if 'ps' in info:
            ps = info['ps']

        if 'add-ps' in info:
            add_ps = info['add-ps']

        if 'bookmarks' in info:
            bookmarks = info['bookmarks']

        if 'title' in info:
            title = info['title']

        if 'author' in info:
            author = info['author']

        if 'verbose' in info:
            verbose = info['verbose']

        if 'silent' in info:
            silent = info['silent']

        if 'no-ps' in info:
            if info['no-ps']:
                meta = False

    if os.path.isfile(output) and not args['--force']:
        if not click.confirm('Overwrite existing "{0:s}"?'.format(output)):
            sys.exit(1)

    try:
        combine(
            files=files,
            output=output,
            openleft=openleft,
            openright=openright,
            meta=meta,
            ps=ps,
            add_ps=add_ps,
            bookmarks=bookmarks,
            title=title,
            author=author,
            verbose=verbose)
    except Exception as e:
        print(str(e))
        sys.exit(1)

    if not silent:
        print('[pdfcombine] {0:s}'.format(output))


if __name__ == '__main__':

    main()




import os
import subprocess
import shutil
import tempfile
from itertools import accumulate

__version__ = '1.1.3'


def run(cmd, verbose=False):
    r'''
Run command, optionally verbose command and output, and return output.
    '''

    if verbose:
        print(cmd)

    out = subprocess.check_output(cmd, shell=True).decode('utf-8')

    if verbose:
        print(out)

    return out


def number_of_pages(files, verbose=False):
    r'''
Read the number of pages of a (list of) PDF(s), using GhostScript.
The output is a (list of) integers.
    '''

    return_int = False

    if type(files) == str:
        files = [files]
        return_int = True

    if verbose:
        print('\n---- reading number of pages per file ----\n')

    n_pages = []

    for file in files:
        cmd = 'gs -q -dNODISPLAY -c "({0:s}) (r) file runpdfbegin pdfpagecount = quit"'.format(file)
        out = run(cmd, verbose)
        n_pages += [int(out)]

    if return_int:
        return n_pages[0]

    return n_pages


def generate_postscript(
        title=None,
        author=None,
        bookmarks=None,
        pages=None,
):
    r'''
Generate PostScript script.

:options:

    **title** (``<str>``)
        PDF title.

    **author** (``<str>``)
        PDF author.

    **bookmarks** (``<list<str>>``)
        List of bookmarks. Length must mage ``pages``.

    **pages** (``<list<int>>``)
        List of page-numbers. Length must mage ``bookmarks``.
    '''

    if type(bookmarks) == str:
        bookmarks = [bookmarks]

    if type(pages) == int:
        pages = [pages]

    out = []

    if title:
        out += ['/Title ({0:s})'.format(title)]

    if author:
        out += ['/Author ({0:s})'.format(author)]

    if bookmarks:
        for bookmark, page in zip(bookmarks, pages):
            out += ['/Page {0:d} /Title ({1:s}) /OUT pdfmark'.format(page, bookmark)]

    if len(out) == 0:
        return ''

    return '[ ' + '\n[ '.join(out)


def combine(
        files,
        output,
        openleft=False,
        openright=False,
        meta=True,
        ps=None,
        add_ps=None,
        bookmarks=True,
        title='Binder',
        author='pdfcombine',
        verbose=False,
):
    r'''
Combine PDFs

:arguments:

    **files** (``<str>`` | ``<list<str>>``)
        List of PDF files to combine.

    **output** (``<str>``)
        Name of output file (overwritten if exists).

:options:

    **openleft** ([``False``] | ``True``)
        Make sure each 'document' begins on a left-page.

    **openright** ([``False``] | ``True``)
        Make sure each 'document' begins on a right-page.

    **meta** ([``True``] | ``False``)
        Write meta-data using a PostScript script
        (see below for options: 'ps', 'add_ps', 'bookmarks', 'title', 'author').

    **ps** (``<str>``)
        If specified the automatically generated PostScript script is overwritten with
        the specified script.

    **add_ps** (``<str>``)
        Append generated/specified PostScript script with the specified script.

    **bookmarks** ([``True``] | ``False`` | ``<list<str>>``)
        If ``True`` the filenames are used as bookmarks in the automatically generated
        PostScript script. One can customise the bookmarks by specifying a list with
        one label per file.

    **title** (``<str>``)
        Specify PDF title. Defaults to "Binder".

    **author** (``<str>``)
        Specify PDF author. Defaults to "pdfcombine".

    **verbose** ([``False``] | ``True``)
        Verbose all commands and their output.
    '''

    temp_dir = None

    if type(files) == str:
        files = [files]

    if type(bookmarks) == str:
        bookmarks = [bookmarks]

    if bookmarks == True:
        bookmarks = [file for file in files]

    if not meta:
        bookmarks = False

    # Basic checking

    if not shutil.which('gs'):
        raise IOError('"gs" not found')

    if openright and openleft:
        raise IOError('"openright" and "openleft" are exclusive options')

    for file in files:
        if not os.path.isfile(file):
            raise IOError('"{0:s}" does not exist'.format(file))

    for file in files:
        if os.path.abspath(file) == os.path.abspath(output):
            raise IOError('"{0:s}" is also an input-file, choose a different output'.format(output))

    # Read number of pages and derive basic information

    if openleft or openright or bookmarks:

        n_pages = number_of_pages(files, verbose)

        is_even = [False if n % 2 != 0 else True for n in n_pages]

    if bookmarks:

        if openleft:
            start_page = [2] + [i if e else i + 1 for i, e in zip(n_pages, is_even)][:-1]
        elif openright:
            start_page = [1] + [i if e else i + 1 for i, e in zip(n_pages, is_even)][:-1]
        else:
            start_page = [1] + [i for i in n_pages]

        start_page = list(accumulate(start_page))

    # PostScript script to set metadata -> write to temporary file

    if meta:

        temp_dir = tempfile.mkdtemp()
        temp_ps = os.path.join(temp_dir, 'pdfcombine.ps')

        if type(ps) != str:
            ps = generate_postscript(
                title=title,
                author=author,
                bookmarks=bookmarks,
                pages=start_page)

        if type(add_ps) == str:
            ps += '\n' + add_ps

        if verbose:
            print('\n---- created "{0:s}" ----\n'.format(temp_ps))
            print(ps)

        open(temp_ps, 'w').write(ps)

    # Construct and run GhostScript command to combine PDFs

    if verbose:
        print('\n---- combining files ----\n')

    cmd = 'gs -sDEVICE=pdfwrite -dBATCH -dNOPAUSE -q -sPAPERSIZE=a4 -o "{0:s}"'.format(output)

    if openleft:
        cmd += ' -c showpage'

    for i, file in enumerate(files):

        if i > 0:
            if (openleft or openright) and not is_even[i-1]:
                cmd += ' -c showpage'

        cmd += ' -f "{0:s}"'.format(file)

    if meta:
        cmd += ' ' + temp_ps

    run(cmd, verbose)

    # Clean-up: remove temporary directory (including its contents)

    if temp_dir is None:
        return

    if verbose:
        print('\n---- cleaning up ----\n')
        print('rm -r {0:s}'.format(temp_dir))
        print('')

    shutil.rmtree(temp_dir)
